# Adapted from https://github.com/crsf-wg/crsf/wiki/Python-Parser

from . import crc
from ctypes import c_uint8 as uint8_t
from typing import Union, Tuple
from enum import IntEnum, Enum
from .elrs_device import ELRSDeviceManager
import hashlib

SYNC_BYTE = 0xC8

class PacketsTypes(IntEnum):
    """
    CRSF Packet Types
    """
    GPS = 0x02
    VARIO = 0x07
    BATTERY_SENSOR = 0x08
    BARO_ALT = 0x09
    HEARTBEAT = 0x0B
    VIDEO_TRANSMITTER = 0x0F
    LINK_STATISTICS = 0x14
    RC_CHANNELS_PACKED = 0x16
    ATTITUDE = 0x1E
    FLIGHT_MODE = 0x21
    DEVICE_INFO = 0x29
    CONFIG_READ = 0x2C
    CONFIG_WRITE = 0x2D
    RADIO_ID = 0x3A
    COMMAND = 0x32
    DEVICE_PING = 0x28

class SubCommandTypes(Enum):
    """
    CRSF Packet Types
    """
    BIND = bytes((0x62, 0x64))
    UNBIND = bytes((0x75, 0x62))
    ID = bytes((0x69, 0x64))
    PWM_CH = bytes((0xF3, ))
    PWM_VAL = bytes((0xF4, ))
    BOOTLOADER = bytes((0x62, 0x6C))


def str2us(microseconds: int) -> Tuple[uint8_t, uint8_t]:
    if microseconds < 256:
        return microseconds, 0
    else:
        return (microseconds >> 8),(microseconds & 0xFF)

def generateUID(phrase):
    uid = [
        int(item) if item.isdigit() else -1
        for item in phrase.split(',')
    ]
    if (4 <= len(uid) <= 6) and all(ele >= 0 and ele < 256 for ele in uid):
        # Extend the UID to 6 bytes, as only 4 are needed to bind
        uid = [0] * (6 - len(uid)) + uid
        uid = bytes(uid)
    else:
        uid = hashlib.md5(("-DMY_BINDING_PHRASE=\""+phrase+"\"").encode()).digest()[0:6]
    return uid

class CRSFCommand:
    """
    Standard interface for sending a crossfire command and reading the response.
    Most routines can be ran easiest by creating the object and calling run() on it.

    Methods
    ----------
    - send: Send the prepared command
    - recv: Receive any response to the command
    - run: (Abstract) Run the command and process any response
    - parse_response

    Attributes
    ----------
    """

    def __init__(self, dev_manager: ELRSDeviceManager = None, cmd: bytearray = None, cmd_type: PacketsTypes = None, subcmd_type: SubCommandTypes = None, payload: bytearray = None) -> None:
        self.cmd: bytearray = cmd
        if cmd is not None:
            self.has_crc = True
        else:
            self.has_crc = False
        if dev_manager is None:
            self.dev_manager = ELRSDeviceManager()
            self.device = self.dev_manager.open_device()
        else:
            self.dev_manager = dev_manager
            if self.dev_manager.dev is not None:
                self.device = self.dev_manager.dev
            else:
                self.device = self.dev_manager.open_device()

        if self.cmd is None:
            if cmd_type is not None:
                self.gen_command_header(cmd_type, subcmd_type)
                if payload is not None:
                    self.cmd.extend(payload)
                self.set_meta()

    def send(self) -> int:
        if self.device is not None and self.cmd is not None:
            if self.dev_manager.dev is None:
                self.dev_manager.open_device()
            try:
                self.dev_manager.dev.write(bytearray(self.cmd))
            except:
                return -1
        else:
            raise Exception("Device and/or command not set!")
        
        return 0

    def recv(self) -> Tuple[PacketsTypes, bytes]:
        data = self.device.read(64)
        ptype, response = self.parse_response(data)
        if ptype != -1:
            return type, response
        else:
            return -1, None

    def parse_response(self, input, expected_pkt: PacketsTypes=None) -> Tuple[PacketsTypes, Union[bytearray, None]]:
        while len(input) > 4:
            val = input[0]
            if val == SYNC_BYTE:
                size = input[1]
                payload = input[2:3 + size]
                if payload[-1] == SYNC_BYTE: payload = payload[:-1]
                packet_crc = payload[-1]
                if packet_crc == crc.calc(payload[:-1]):
                    packet_type: PacketsTypes = PacketsTypes(input[2])
                    if expected_pkt is None or expected_pkt == packet_type:
                        return (packet_type, payload[3:-1])
            input = input[1:]
        return (-1, None)
    
    def gen_command_header(self, cmd_type : PacketsTypes, subcmd_type: SubCommandTypes = None):
        data_out = bytearray()
        data_out.append(0xEC)  # ADDR -> RECEIVER
        data_out.append(0x00)
        data_out.append(cmd_type.value)
        if cmd_type == PacketsTypes.COMMAND:
            if subcmd_type is None:
                raise ValueError("No subcommand type!")
            data_out.extend(subcmd_type.value)
        self.cmd = data_out


    def add_crc(self) -> None:
        self.cmd.append(crc.calc(self.cmd[2:]))
        self.has_crc = True

    def set_len(self) -> None:
        self.cmd[1] = len(self.cmd) - 2 if self.has_crc else len(self.cmd) - 1

    def set_meta(self) -> None:
        self.set_len()
        self.add_crc()

    def run(self):
        raise NotImplementedError()


class CRSFBind(CRSFCommand):
    """
    Put the receiver into binding mode
    """

    def __init__(self, dev_manager=None) -> None:
        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.COMMAND,
            subcmd_type=SubCommandTypes.BIND)
    
    def recv(self) -> bytes:
        return None
    
    def run(self) -> int:
        return self.send()


class CRSFDev(CRSFCommand):

    def __init__(self, dev_manager=None) -> None:
        payload = bytes((0xEC, 0xC8, 0xEC))
        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.DEVICE_PING,
            payload=payload)

    def recv(self) -> Tuple[PacketsTypes, bytes]:
        ptype = -1
        count = 3
        response = None
        while ptype == -1 and count > 0:
            data = self.device.read(64)
            count -= 1
            ptype, response = self.parse_response(data, expected_pkt=PacketsTypes.DEVICE_INFO)
        if ptype != -1:
            return response
        else:
            return None

    def parse_info(self, response) -> dict:
        _device_name:str =""
        _sw_version:str =""

        for b in response:
            if b == 0:
                break
            _device_name += chr(b)
    
        # Get Version -> Version is always the [-7:-3] bytes in the packet we receive in the current version of ELRS
        _sw_version = '.'.join([str(int(b)) for b in response[-6:-2]])
        if _sw_version[0] == "0":
            _sw_version = _sw_version[2:]
        _hw_version = str(response[-10:-6]) if int.from_bytes(response[-10:-6], "big") != 0 else "ELRS"
        _serial_num = str(response[-14:-10], "utf-8")
        if _serial_num != 'ELRS':
            print(f"Failed to validate version info: {_serial_num} != \"ELRS\"")

        return {
            "name": _device_name,
            "sw_ver": _sw_version,
            "hw_ver": _hw_version,
            "serial_num": _serial_num,
            "modal_elrs": not _hw_version == "ELRS", # True if _hw_version is set to m184 or other device
        }

    def format_info(self, info):
        return f"{info['name']}:{info['sw_ver']}"
        

    def run(self, retries=5) -> dict:
        for _ in range(retries):
            self.send()
            response = self.recv()
            if response is not None:
                break
        if response is not None:
            return self.parse_info(response)
        else:
            return None


class CRSFUnbind(CRSFCommand):
    """
    Unbind the receiver
    """

    def __init__(self, dev_manager=None) -> None:
        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.COMMAND,
            subcmd_type=SubCommandTypes.UNBIND)
    
    def recv(self) -> bytes:
        return None
    
    def run(self) -> int:
        return self.send()


class CRSFSetID(CRSFCommand):
    """
    Sets the reciever's UID based on a binding phrase or directly
    """

    def __init__(self, binding_phrase : str = None, new_id : bytes = None, dev_manager=None, cmd=None) -> None:
        if new_id is not None:
            payload = new_id
        elif binding_phrase is not None:
            payload = generateUID(binding_phrase)
        else:
            raise ValueError("No binding phrase or id")

        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.COMMAND,
            subcmd_type=SubCommandTypes.ID,
            payload=payload)
    
    def recv(self) -> bytes:
        return None
    
    def run(self) -> int:
        return self.send()


class CRSFSetPWMCh(CRSFCommand):
    """
    Set the input channel a pwm pin should listen on
    """

    def __init__(self, pin: int, in_ch: int, dev_manager=None, cmd=None) -> None:
        payload = bytes((pin + 1, in_ch))
        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.COMMAND,
            subcmd_type=SubCommandTypes.PWM_CH,
            payload=payload)
    
    def recv(self) -> bytes:
        return None
    
    def run(self) -> int:
        return self.send()


class CRSFSetPWMVal(CRSFCommand):
    """
    Set the value of an input channel for the pwm pins
    """

    def __init__(self, in_ch: int, microseconds: int = None, duty: int = None, dev_manager=None, cmd=None) -> None:
        if duty is not None:
            payload = bytearray((in_ch, 0x64, 0x00, duty))
        else:
            payload = bytearray((in_ch, 0x73))
            payload.extend(str2us(microseconds))

        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.COMMAND,
            subcmd_type=SubCommandTypes.PWM_VAL,
            payload=payload)
            
    def recv(self) -> bytes:
        return None
    
    def run(self) -> int:
        return self.send()


class CRSFSetBootloader(CRSFCommand):
    """
    Jump back to the bootloader for restarting the device or uploading firmware
    """

    def __init__(self, dev_manager=None) -> None:
        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.COMMAND,
            subcmd_type=SubCommandTypes.BOOTLOADER)
    
    def recv(self) -> bytes:
        return None
    
    def run(self) -> int:
        return self.send()


def send_tx_bind():
    dev = ELRSDeviceManager(baud = 400000)
    cmd_payload = bytearray((0xC8, 0x06, 0x32, 0xEE, 0xEA, 0x10, 0x01))
    cmd = CRSFCommand(dev_manager=dev, cmd=cmd_payload)
    cmd.add_crc()
    print(cmd.cmd.hex())
    cmd.send()

if __name__ == "__main__":
    #dev = CRSFSetID(binding_phrase="modal dbg phrase")
    dev = ELRSDeviceManager(baud = 400000)
    cmd_payload = bytearray((0xC8, 0x00, 0x32, 0xEE, 0xEA, 0x10, 0x01))
    cmd = CRSFCommand(dev_manager=dev, cmd=cmd_payload)
    cmd.set_meta()
    cmd.send()
