import os.path
from .elrs_device import ELRSDeviceManager
from .voxl_utils import is_voxl
from enum import Enum
from glob import glob
import json
import subprocess
from .crsf import CRSFSetBootloader, CRSFDev
from time import sleep
from typing import List
import pdb
import os
import re
from subprocess import PIPE

class PlatformTypes(Enum):
    ModalAI_STM32 = 1,
    ESP = 2,

lua_name_map = {
    "MODALAI_M0184_RX_via_UART": [{
            "manufacturer": "ModalAI",
            "product_name": "M0139",
            "lua_name": "ModalAI M0139",
            "platform": PlatformTypes.ModalAI_STM32
        },
        {
            "manufacturer": "ModalAI",
            "product_name": "M0184",
            "lua_name": "ModalAI M0184",
            "platform": PlatformTypes.ModalAI_STM32
        }],
    "MODALAI_M0184_HWIL_RX_via_UART": [{
            "manufacturer": "ModalAI",
            "product_name": "M0139_DEBUG",
            "lua_name": "M0139 HWIL",
            "platform": PlatformTypes.ModalAI_STM32
        },
        {
            "manufacturer": "ModalAI",
            "product_name": "M0184_DEBUG",
            "lua_name": "M0184 HWIL",
            "platform": PlatformTypes.ModalAI_STM32
        }],
    "MODALAI_M0193_RX_via_UART": [{
            "manufacturer": "ModalAI",
            "product_name": "M0193",
            "lua_name": "M0193 HWIL",
            "platform": PlatformTypes.ModalAI_STM32
        }],
    "BETAFPV_900_RX_via_UART": [
        {
            "manufacturer": "betafpv",
            "product_name": "BETAFPV 900MHz RX",
            "lua_name": "BETAFPV 900RX",
            "platform": PlatformTypes.ESP
        }],
    "FrSky_R9Mini_via_UART": [
        {
            "manufacturer": "frsky",
            "product_name": "FrSky R9 Mini",
            "lua_name": "FrSky R9MM/Mini",
            "platform": PlatformTypes.ModalAI_STM32
        },
        {
            "manufacturer": "frsky",
            "product_name": "FrSky R9 Mini",
            "lua_name": "FrSky R9MM",
            "platform": PlatformTypes.ModalAI_STM32
        }],
}



def version_tuple(v):
   filled = []
   for point in v.split("."):
      filled.append(point.zfill(8))
   return tuple(filled)

class ELRSFirmwareManager():
    def __init__(self, dev: ELRSDeviceManager) -> None:
        self.dev = dev
        self.firmware_dir = self.dev.FIRMWARE_DIR

    def get_unified_targets(self, targets_json: dict, fw_list: List[str], version: str, local: bool) -> List[dict]:
        targets_list = []
        for manufacturer in targets_json.keys():
            for radio_type in targets_json[manufacturer].keys():
                if radio_type != "name":
                    for radio_ver in targets_json[manufacturer][radio_type].keys():
                        for fw in fw_list:
                            if targets_json[manufacturer][radio_type][radio_ver]["firmware"] in fw:
                                targets_list.append({
                                    "manufacturer": manufacturer,
                                    "product_name": targets_json[manufacturer][radio_type][radio_ver]["product_name"],
                                    "lua_name": targets_json[manufacturer][radio_type][radio_ver]["lua_name"],
                                    "layout_file": targets_json[manufacturer][radio_type][radio_ver]["layout_file"],
                                    "version": version,
                                    "firmware": fw,
                                    "local": local,
                                    "chip": targets_json[manufacturer][radio_type][radio_ver]["platform"],
                                    "platform": PlatformTypes.ESP,
                                })
        return targets_list

    def get_stm_targets(self, fw: str, version: str, local: bool) -> List[dict]:
        targets = []
        for fw_desc in lua_name_map[fw]:
            new_target = fw_desc.copy()
            new_target["layout_file"] = None
            new_target["version"] = version
            new_target["firmware"] = fw
            new_target["local"] = local
            #new_target["platform"] = PlatformTypes.ModalAI_STM32
            new_target["chip"] = "stm32"
            targets.append(new_target)
        return targets

    def get_local_targets(self, version_path: str, platform: PlatformTypes = None):
        _, version = os.path.split(version_path)
        fw_list = os.listdir(version_path)
        target_json = None
        targets = []

        # The hardware folder contains information about the unified targets
        if "hardware" in fw_list:
            fw_list.remove("hardware")
            if platform is None or platform == PlatformTypes.ESP:
                target_json = os.path.join(version_path, "hardware/targets.json")
                with open(target_json, "r") as target_f:
                    targets.extend(self.get_unified_targets(json.load(target_f), fw_list, version, True))

        # All other targets have their own directories
        for fw in fw_list:
            if "Unified" not in fw and fw in lua_name_map and (platform is None or platform != PlatformTypes.ESP):
                targets.extend(self.get_stm_targets(fw, version, True))

        return targets
    
    def local_images(self, name: str = None, platform: PlatformTypes = None, version: str = None, root_path: str = None) -> List[dict]:
        if root_path is None:
            root_path = self.firmware_dir
        available_versions = os.listdir(root_path)
        
        # version specifier
        if version is not None:
            available_versions = [x for x in available_versions if x[0:len(version)] == version]
        
        available_versions = [x for x in available_versions if re.match("[0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+", x) is not None]

        if len(available_versions) == 0:
            return []

        targets = []
        [targets.extend(self.get_local_targets(os.path.join(root_path, x), platform=platform)) for x in available_versions]

        if name is not None:
            targets = [x for x in targets if x["lua_name"] == name]
        
        return targets

    def get_cloud_targets(self, name: str = None, platform: PlatformTypes = None, version: str = None, root_path: str = None):
        self.dev.FIRMWARE_DIR


    def available_images(self, local=None, name: str = None, platform: PlatformTypes = None, version: str = None) -> list:
        return self.local_images(name, platform, version)
    
    def select_image(self, targets: List[dict], min_version: str = None):
        version_list = [t["version"] for t in targets]
        if min_version is not None:
            version_list.append(min_version)
        max_ver = max([version_tuple(x) for x in version_list])
        max_list = [t for t in targets if version_tuple(t["version"]) == max_ver]
        if len(max_list) == 1:
            return max_list
        elif len(max_list) == 0:
            return None
        
        # Prefer local images if multiple images exist
        local_images = False
        cloud_images = False
        for t in max_list:
            if t["local"]:
                local_images = True
            else:
                cloud_images = True
        if local_images and cloud_images:
            max_list = [t for t in targets if t["local"]]

        if len(max_list) == 0:
            return None
        return max_list

    def get_firmware(self, local=None, name: str = None, platform: PlatformTypes = None, version: str = None, min_version: str = None):
        targets = self.available_images(local, name, platform, version)
        final_images = self.select_image(targets)
        if final_images is None:
            print("Error parsing firmware versions")
            return None
        if len(final_images) > 1:
            print("Ambiguous firmware files, manual selection necessary!")
            return None
        else:
            print(f"Selected firmware image:\n{ELRSFirmwareManager.fw_string(final_images[0])}")
            return final_images[0]
        
    def get_firmware_file(self, fw: dict):
        root_path = self.firmware_dir
        if fw["local"]:
            try:
                fw_folder = os.path.join(os.path.join(root_path, fw["version"]), fw["firmware"])
                fw_file = glob(os.path.join(fw_folder, "*.bin"))
                return fw_file[0]
            except:
                return None
        else:
            raise NotImplementedError("Remote files not yet implemented")
        
    def upload_firmware_from_dict(self, fw_dict: dict):
        return self.upload_firmware(self.get_firmware_file(fw_dict), fw_dict["platform"], fw_dict["chip"])

    def upload_firmware(self, firmware_file: str, platform: PlatformTypes = PlatformTypes.ModalAI_STM32, chip = None):
        upload_dev = self.dev.device
        if upload_dev is None:
            if is_voxl():
                upload_dev = "/dev/slpi-uart-7"
            else:
                upload_dev = "/dev/ttyUSB0"
        elif "slpi" in upload_dev:
            upload_dev = "/dev/slpi-uart-7"
        if platform == PlatformTypes.ModalAI_STM32:
            # Reset into bootloader
            CRSFSetBootloader(self.dev).run()
            self.dev.close(False, close_fds=True)
            print("Flashing...") 
            print("If you do not see a blinking green led (indicating that the firmware is being uploaded), please power cycle the receiver now")
            # Flash receiver 
            print(f"Uploading firmware file: {firmware_file}")
            os.chdir("/usr/share/modalai/voxl-esc-tools/")
            proc = subprocess.Popen(["python3","/usr/share/modalai/voxl-esc-tools/voxl-esc-upload-firmware.py", 
                                    "--firmware-file", firmware_file, 
                                    "--id", "0",
                                    "--skip-firmware-reset", "1",
                                    "--device", upload_dev], stderr=subprocess.DEVNULL)
            proc.wait(timeout=240)
            if (is_voxl()):
                #######################################
                ## TODO: Remove this, libvoxl_io currently doesn't clean up fds correctly
                #######################################
                # simply exit after uploading any firmware since opening up the uart again seems to always fail
                exit()
                os.closerange(3, 10)
                #subprocess.run("voxl-reset-slpi")

            self.dev.open_device()
            return 0
        elif platform == PlatformTypes.ESP:
            # Reset into bootloader
            CRSFSetBootloader(self.dev).run()
            self.dev.close(False)
            if is_voxl():
                subprocess.run("voxl-reset-slpi")
            sleep(5)
            print("Flashing... Please do not power off the receiver")
            if chip == "esp8285":
                chip = "esp8266"

            import esptool

            # Flash receiver 
            rx_flash_command = [
                "python3", "-m", "esptool",
                #"--baud", "460800",
                "--port", str(upload_dev),
                #"--chip", str(chip),
                "--chip", "auto",
                "--before", "no_reset",
                "--after", "soft_reset",
                "write_flash",
                "0x00",
                str(firmware_file)
            ]
            #result = esptool.main(rx_flash_command)
            proc = subprocess.Popen(rx_flash_command, stderr=subprocess.DEVNULL)
            proc.wait(timeout=180)
            if is_voxl():
                subprocess.run("voxl-reset-slpi")
            sleep(5)

            self.dev.open_device()
            return 0
            
    def auto_update(self, name: str, min_ver: str = None):
        ret = -1
        fw = self.get_firmware(name=name, min_version=min_ver)
        if fw is not None:
            if min_ver is not None and fw["version"] == min_ver:
                print(f"ELRS Receiver \"{name}\" already up to date!")
                return 0
            fw_file = self.get_firmware_file(fw)

            if fw_file is not None:
                ret = self.upload_firmware(fw_file, fw["platform"], fw["chip"])

            if ret == 0:
                print("Successfuly flashed device")
                sleep(1)
                new_ver = "0"
                new_ver_info = CRSFDev(self.dev).run()
                if(new_ver_info is not None):
                    new_ver = new_ver_info["sw_ver"]
                if fw["version"] == new_ver:
                    return 0
                else:
                    print("Incorrect ELRS version identified")
                    return -1
        return ret


    def fw_string(fw: dict):
        return f"{fw['manufacturer']} : {fw['product_name']} | ELRS v{fw['version']} | Local: {fw['local']}"
